var http  = require("http");
var fs    = require("fs");
var wsAPI = require("websocket.io");
var path  = require("path");
var url   = require("url");

var Server = function(port, httpRoot, livereload) {
	var self      = this;
	
	this.contentTypes = {
		".js"		: "text/javascript",
		".htc"		: "text/x-component",
		".css"		: "text/css",
		".png"		: "image/png",
		".gif"		: "image/gif",
		".jpg"		: "image/jpeg",
		"default"	: "text/html"
	};
	
	this.customRequests = {
		"/makatto-reload.js" : this._customRequestReload
	};
	
	this.port     = port;
	this.httpRoot = httpRoot;
	this.livereload    = !!livereload;
	
	this.server   = this._createServer();
	this.ws       = this._createWS();
	this.sockets = [];
};

Server.prototype = {
	
	start : function(callback) {
		var self = this;
		
		this.server.listen(this.port, function() {
			callback && callback();
		});
	},
	
	stop : function(callback) {
		for (var i = 0; i < this.sockets.length; i++) {
			this.sockets[i] && this.sockets[i].close();
		}
		this.server.close(callback);
	},
	
	_createServer : function() {
		var self = this;
		return http.createServer(function(request, response) {
			self._onRequest.apply(self, arguments);
		});
	},
	
	_createWS : function() {
		var self = this;
		
		var ws = wsAPI.attach(this.server);
		ws.on("connection", function(_socket) {
			self._onWsConnect.apply(self, arguments);
		});
	},
	
	_onRequest : function(request, response) {
		var self = this;
		var filePath;
		var urlParts = url.parse(request.url, true);
		
		if (urlParts.pathname === "/favicon.ico") {
			return;
		}
		
		if (/.*\/$/.test(urlParts.pathname)) { // ends with slash
			// check if index.html is present
			if (fs.existsSync(self.httpRoot + "/" + urlParts.pathname + "/index.html")) {
				response.writeHead(301, {
					"Location" : urlParts.pathname + "index.html"
				});
				response.end();
				console.log("301 - / -> /index.html");
				return;
			}
			// show indexes otherwise
			else {
				// if dir exists, add a slash and redirect
				if (fs.existsSync(self.httpRoot + "/" + urlParts.pathname)) {
					response.writeHead(200, {
						"Content-Type" : "text/html;charset=UTF-8",
						"Cache-Control" : "no-cache",
						"Expires" : "-1"
					});
					response.end(self._indexes(urlParts), "utf-8");
					return;
				}
				// 404 otherwise
				else {
					response.writeHead(404);
					response.end(self._notFound(urlParts), "utf-8");
					console.log("404 - " + request.url);
					return;
				}
			}
		}
		else {
			if (/.*\..+$/.test(urlParts.pathname)) { // has extension
				if (self.customRequests[urlParts.pathname]) {
					self.customRequests[urlParts.pathname].call(self, urlParts, request, response);
					return;
				}
				else {
					filePath = self.httpRoot + "/" + urlParts.pathname;
				}
			}
			else { // does not have extension and does not end with a slash
				// if dir exists, add a slash and redirect
				if (fs.existsSync(self.httpRoot + "/" + urlParts.pathname)) {
					if (fs.statSync(self.httpRoot + "/" + urlParts.pathname).isDirectory()) {
						response.writeHead(301, {
							"Location" : urlParts.pathname + "/"
						});
						response.end();
						console.log("301 - -> /");
						return; 
					}
					else {
						filePath = self.httpRoot + "/" + urlParts.pathname;
					}
				}
				// 404 otherwise
				else {
					response.writeHead(404);
					response.end(self._notFound(urlParts), "utf-8");
					console.log("404 - " + request.url);
					return;
				}
			}
		}
		
		fs.exists(filePath, function(exists) {
			var ext = path.extname(filePath);
			
			if (!exists) {
				response.writeHead(404);
				response.end(self._notFound(urlParts), "utf-8");
				console.log("404 - " + request.url);
				return;
			}
			fs.readFile(filePath, function(error, content) {
				if (error) {
					response.writeHead(500);
					response.end("500", "utf-8");
					console.log("500 - " + request.url);
					return;
				}
				
				var contentS = content.toString();
				
				if (ext === ".html" && self.livereload === true) {
					if (contentS.indexOf("</head>") > 0) {
						content = contentS.replace("<\/head>", "\t<script src=\"/makatto-reload.js\"><\/script>\n<\/head>");
					}
					else if (contentS.indexOf("</body>") > 0) {
						content = contentS.replace("<\/body>", "\t<script src=\"/makatto-reload.js\"><\/script>\n<\/body>");
					}
					else {
						content += "\n<script src=\"/makatto-reload.js\"><\/script>";
					}
				}
				
				response.writeHead(200, {
					"Content-Type" : self.contentTypes[ext] || self.contentTypes["default"] + ";charset=UTF-8",
					"Cache-Control" : "no-cache",
					"Expires" : "-1"
				});
				response.end(content, "utf-8");
				// console.log("200 - " + request.url);
			});
		});
	},
	
	_customRequestReload : function(urlParts, request, response) {
		var script = fs.readFileSync(__dirname + "/reload.js", "utf-8");
		
		response.writeHead(200, {
			"Content-Type" : "text/javascript;charset=UTF-8",
			"Cache-Control" : "no-cache",
			"Expires" : "-1"
		});
		response.end(script, "utf-8");
		console.log("200 - " + request.url);
	},
	
	_onWsConnect : function(_socket) {
		var self = this;
		
		this.sockets.push(_socket);
		_socket.on("message", function(msg) {
			self._onWsMessage.apply(self, arguments);
		});
	},
	
	_onWsMessage : function(msg) {
		console.log("WS< " + msg);
	},
	
	_notFound : function(urlParts) {
		return this._renderTemplate(global.config.binDir, "404-template.html", {
			"title"      : "404 - Not found",
			"url"        : urlParts.pathname
		});
	},
	
	_indexes : function(urlParts) {
		var dir    = global.config.distDir + urlParts.pathname;
		var pages  = [];
		
		if (fs.existsSync(dir)) {
			fs.readdirSync(dir).forEach(function(page) {
				pages.push({filename : page, isDir : fs.statSync(dir + "/" + page).isDirectory()});
			});
		}
		
		return this._renderTemplate(global.config.binDir, "index-template.html", {
			"title"      : urlParts.pathname,
			"isRoot"     : urlParts.path === "/",
			"pages"      : pages
		});
	},
	
	_renderTemplate : function(templateRoot, templatePath, templateData) {
		var swig   = require("swig");
		var tpl;
		
		swig.init({root : templateRoot});
		tpl = swig.compileFile(templateRoot + "/" + templatePath);
		
		return tpl.render(templateData);
	}
	
};

exports.Server = Server;
